import sphinx


def skip_member(app, what, name, obj, skip, options):
    # skip special members except for __init__ and __call__
    return sphinx.ext.napoleon._skip_member(app, what, name, obj, skip, options) or \
           name[:2] == '__' and name not in ['__init__', '__call__', '__next__', '__iter__']


def setup(app):
    # wrap napoleon's skip-member
    app.connect('autodoc-skip-member', skip_member)
    return {'version': sphinx.__display_version__}



from unittest import TestCase

import Orange

import orangecontrib.conformal as cp


class TestClassification(TestCase):
    def setUp(self):
        self.tab = Orange.data.Table('iris')
        self.inst, self.tab = self.tab[0], self.tab[1:]
        self.train, self.test = self.tab[0::2], self.tab[1::2]
        self.train, self.calibrate = self.train[0::2], self.train[1::2]
        self.nc = cp.nonconformity.InverseProbability(Orange.classification.NaiveBayesLearner())

    def test_transductive(self):
        predictor = cp.classification.TransductiveClassifier(self.nc, self.train)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_inductive(self):
        predictor = cp.classification.InductiveClassifier(self.nc, self.train, self.calibrate)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_cross(self):
        predictor = cp.classification.CrossClassifier(self.nc, 4, self.train)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_loo(self):
        predictor = cp.classification.LOOClassifier(self.nc, self.train)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_nc_type(self):
        nc_class = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        nc_regr = cp.nonconformity.AbsError(Orange.regression.LinearRegressionLearner())
        cp.classification.TransductiveClassifier(nc_class)
        self.assertRaises(AssertionError, cp.classification.TransductiveClassifier, nc_regr)
        cp.classification.InductiveClassifier(nc_class)
        self.assertRaises(AssertionError, cp.classification.InductiveClassifier, nc_regr)
        cp.classification.CrossClassifier(nc_class, 5)
        self.assertRaises(AssertionError, cp.classification.CrossClassifier, nc_regr, 5)

    def test_str(self):
        nc = cp.nonconformity.InverseProbability(Orange.classification.NaiveBayesLearner())
        predictor = cp.classification.InductiveClassifier(nc)
        s = str(predictor)
        self.assertIn("InductiveClassifier", s)
        self.assertIn("InverseProbability", s)


class TestRegression(TestCase):
    def setUp(self):
        self.tab = Orange.data.Table('housing')
        self.inst, self.tab = self.tab[1], self.tab[2:]
        self.train, self.test = self.tab[0::2], self.tab[1::2]
        self.train, self.calibrate = self.train[0::2], self.train[1::2]
        self.nc = cp.nonconformity.AbsError(Orange.regression.LinearRegressionLearner())

    def test_inductive(self):
        predictor = cp.regression.InductiveRegressor(self.nc, self.train, self.calibrate)
        lo, hi = predictor(self.inst, 0.2)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

    def test_cross(self):
        predictor = cp.regression.CrossRegressor(self.nc, 4, self.train)
        lo, hi = predictor(self.inst, 0.2)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

    def test_loo(self):
        predictor = cp.regression.LOORegressor(self.nc, self.train)
        lo, hi = predictor(self.inst, 0.2)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

    def test_nc_type(self):
        nc_class = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        nc_regr = cp.nonconformity.AbsError(Orange.regression.LinearRegressionLearner())
        cp.regression.InductiveRegressor(nc_regr)
        self.assertRaises(AssertionError, cp.regression.InductiveRegressor, nc_class)
        cp.regression.CrossRegressor(nc_regr, 5)
        self.assertRaises(AssertionError, cp.regression.CrossRegressor, nc_class, 5)

    def test_str(self):
        nc = cp.nonconformity.AbsError(Orange.regression.LinearRegressionLearner())
        predictor = cp.regression.InductiveRegressor(nc)
        s = str(predictor)
        self.assertIn("InductiveRegressor", s)
        self.assertIn("AbsError", s)


class TestMondrian(TestCase):
    def setUp(self):
        self.tab = Orange.data.Table('iris')
        self.inst, self.tab = self.tab[0], self.tab[1:]
        self.train, self.test = self.tab[0::2], self.tab[1::2]
        self.train, self.calibrate = self.train[0::2], self.train[1::2]
        self.nc = cp.nonconformity.InverseProbability(Orange.classification.NaiveBayesLearner())

    def test_transductive(self):
        predictor = cp.classification.TransductiveClassifier(self.nc, self.train, mondrian=True)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_inductive(self):
        predictor = cp.classification.InductiveClassifier(self.nc, self.train, self.calibrate, mondrian=True)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_cross(self):
        predictor = cp.classification.CrossClassifier(self.nc, 7, self.train, mondrian=True)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_loo(self):
        predictor = cp.classification.LOOClassifier(self.nc, self.train, mondrian=True)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])


class TestEfficiency(TestCase):
    def test_individual_classification(self):
        tab = Orange.data.Table('iris')
        train, test = tab[:123], tab[123]  # borderline case
        nc = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        pred = cp.classification.CrossClassifier(nc, 3, train)
        p = pred.predict(test)
        cred, conf = p.credibility(), p.confidence()
        self.assertLess(cred, 0.5)
        d = 1e-6
        self.assertEqual(len(pred(test, 1-(conf-d))), 1)
        self.assertGreater(len(pred(test, 1-(conf+d))), 1)
        self.assertEqual(len(pred(test, (cred+d))), 0)
        self.assertGreater(len(pred(test, (cred-d))), 0)



from unittest import TestCase

import sklearn

import Orange

import orangecontrib.conformal as cp


class TestClassification(TestCase):
    def setUp(self):
        self.tab = Orange.data.Table('iris')
        self.inst, self.tab = self.tab[0], self.tab[1:]
        self.train, self.calibrate = self.tab[0::2], self.tab[1::2]

    def test_Models(self):
        nc = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
        predictor = cp.classification.InductiveClassifier(nc, self.train, self.calibrate)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

        nc = cp.nonconformity.ProbabilityMargin(Orange.classification.LogisticRegressionLearner())
        predictor = cp.classification.InductiveClassifier(nc, self.train, self.calibrate)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_SVM(self):
        tab = Orange.data.Table('titanic')
        inst, tab = tab[0], tab[1:]
        train, calibrate = tab[:1500], tab[1500:]

        nc = cp.nonconformity.SVMDistance(sklearn.svm.SVC())
        predictor = cp.classification.InductiveClassifier(nc, train, calibrate)
        self.assertListEqual(predictor(inst, 0.1), ['no'])

    def test_Neighbours(self):
        nc = cp.nonconformity.KNNDistance(Orange.distance.Euclidean(), 10)
        predictor = cp.classification.InductiveClassifier(nc, self.train, self.calibrate)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

        nc = cp.nonconformity.KNNFraction(Orange.distance.Euclidean(), 10, weighted=True)
        predictor = cp.classification.InductiveClassifier(nc, self.train, self.calibrate)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_Mahalanobis(self):
        nc = cp.nonconformity.KNNDistance(Orange.distance.Mahalanobis(), 5)
        predictor = cp.classification.InductiveClassifier(nc, self.train, self.calibrate)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

        Mah = Orange.distance.Mahalanobis().fit(self.train)
        nc = cp.nonconformity.KNNDistance(Mah, 5)
        predictor = cp.classification.InductiveClassifier(nc, self.train, self.calibrate)
        self.assertListEqual(predictor(self.inst, 0.1), ['Iris-setosa'])

    def test_str(self):
        self.assertIn("InverseProbability",
                      str(cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())))
        self.assertIn("ProbabilityMargin",
                      str(cp.nonconformity.ProbabilityMargin(Orange.classification.LogisticRegressionLearner())))
        self.assertIn("SVMDistance",
                      str(cp.nonconformity.SVMDistance(sklearn.svm.SVC())))
        self.assertIn("KNNDistance",
                      str(cp.nonconformity.KNNDistance(Orange.distance.Euclidean(), 10)))
        self.assertIn("KNNFraction",
                      str(cp.nonconformity.KNNFraction(Orange.distance.Euclidean(), 10)))
        self.assertIn("LOOClassNC",
                      str(cp.nonconformity.LOOClassNC(Orange.classification.LogisticRegressionLearner(),
                                                      Orange.distance.Euclidean(), 10)))


class TestRegression(TestCase):
    def setUp(self):
        self.tab = Orange.data.Table('housing')[:100]
        self.inst, self.tab = self.tab[1], self.tab[2:]
        self.train, self.calibrate = self.tab[:70], self.tab[70:]

    def test_Models(self):
        nc = cp.nonconformity.AbsError(Orange.regression.LinearRegressionLearner())
        predictor = cp.regression.InductiveRegressor(nc, self.train, self.calibrate)
        lo, hi = predictor(self.inst, 0.1)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

        nc = cp.nonconformity.AbsErrorNormalized(Orange.regression.LinearRegressionLearner(),
                                                 Orange.distance.Euclidean(), 10)
        predictor = cp.regression.InductiveRegressor(nc, self.train, self.calibrate)
        lo, hi = predictor(self.inst, 0.1)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

    def test_RandomForest(self):
        nc = cp.nonconformity.AbsErrorRF(Orange.regression.RandomForestRegressionLearner(),
                                         sklearn.ensemble.RandomForestRegressor())
        predictor = cp.regression.InductiveRegressor(nc, self.train, self.calibrate)
        lo, hi = predictor(self.inst, 0.1)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

    def test_ErrorModel(self):
        nc = cp.nonconformity.ErrorModelNC(Orange.regression.SVRLearner(),
                                           Orange.regression.LinearRegressionLearner())
        predictor = cp.regression.InductiveRegressor(nc, self.train, self.calibrate)
        lo, hi = predictor(self.inst, 0.1)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

    def test_Neighbours(self):
        nc = cp.nonconformity.AbsErrorKNN(k=10)
        self.assertIsInstance(nc.distance, Orange.distance.Euclidean)
        predictor = cp.regression.InductiveRegressor(nc, self.train, self.calibrate)
        lo, hi = predictor(self.inst, 0.1)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

        nc = cp.nonconformity.AvgErrorKNN(Orange.distance.Euclidean(), 10)
        predictor = cp.regression.InductiveRegressor(nc, self.train, self.calibrate)
        lo, hi = predictor(self.inst, 0.1)
        self.assertTrue(lo <= float(self.inst.y) <= hi)

    def test_str(self):
        self.assertIn("AbsError",
                      str(cp.nonconformity.AbsError(Orange.regression.LinearRegressionLearner())))
        self.assertIn("AbsErrorRF",
                      str(cp.nonconformity.AbsErrorRF(Orange.regression.LinearRegressionLearner(),
                                                      sklearn.ensemble.RandomForestRegressor())))
        self.assertIn("ErrorModelNC",
                      str(cp.nonconformity.ErrorModelNC(Orange.regression.SVRLearner(),
                                                        Orange.regression.LinearRegressionLearner())))
        self.assertIn("AbsErrorNormalized",
                      str(cp.nonconformity.AbsErrorNormalized(Orange.regression.LinearRegressionLearner(),
                                                              Orange.distance.Euclidean(), 10)))
        self.assertIn("LOORegrNC",
                      str(cp.nonconformity.LOORegrNC(Orange.regression.LinearRegressionLearner(),
                                                     Orange.distance.Euclidean(), 10)))
        self.assertIn("AbsErrorKNN",
                      str(cp.nonconformity.AbsErrorKNN(Orange.distance.Euclidean(), 10)))
        self.assertIn("AvgErrorKNN",
                      str(cp.nonconformity.AvgErrorKNN(Orange.distance.Euclidean(), 10)))



"""Classification module contains methods for conformal classification.

Conformal classifiers predict a set of classes (not always a single class) under a given
significance level (error rate). Every classifier works in combination with a nonconformity measure
and on average predicts the correct class with the given error rate. Lower error rates result in
smaller sets of predicted classes.

Structure:

- ConformalClassifier
    - Transductive (:py:class:`TransductiveClassifier`)
    - Inductive (:py:class:`InductiveClassifier`)
    - Cross (:py:class:`CrossClassifier`)
"""

from copy import deepcopy

import numpy as np
from sklearn.model_selection import KFold

from Orange.data import Table, Instance

from orangecontrib.conformal.base import ConformalPredictor
from orangecontrib.conformal.nonconformity import ClassNC


class PredictionClass:
    """Conformal classification prediction object,
    which is produced by the :py:func:`ConformalClassifier.predict` method.

    Attributes:
        p (List): List of pairs (p-value, class)
        eps (float): Default significance level (error rate).

    Examples:
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> tcp = TransductiveClassifier(InverseProbability(NaiveBayesLearner()), train)

        >>> prediction = tcp.predict(test[0], 0.1)
        >>> print(prediction.confidence(), prediction.credibility())

        >>> prediction = tcp.predict(test[0])
        >>> print(prediction.classes(0.1), prediction.classes(0.9))
    """

    def __init__(self, p, eps):
        """Initialize the prediction.

        Args:
            p (List): List of pairs (p-value, class)
            eps (float): Default significance level (error rate).
        """
        self.p = p
        self.eps = eps

    def classes(self, eps=None):
        """ Compute the set of classes under the default or given `eps` value.

        Args:
            eps (float): Significance level (error rate).

        Returns:
            List of predicted classes.
        """
        if eps is None:
            eps = self.eps
            assert(eps is not None)
        return [y for p_y, y in self.p if p_y > eps]

    def verdict(self, ref, eps=None):
        """Conformal classification prediction is correct when the actual class appears
        among the predicted classes.

        Args:
            ref: Reference/actual class
            eps (float): Significance level (error rate).

        Returns:
            True if the prediction with default or specified `eps` is correct.
        """
        return ref in self.classes(eps)

    def confidence(self):
        """Confidence is an efficiency measure of a single prediction.

        Computes minimum :math:`\\mathit{eps}` that would still result in a prediction of a single label.
        :math:`\\mathit{eps} = \\text{second\_largest}(p_i)`

        Returns:
            float: Confidence :math:`1-\\mathit{eps}`.
        """
        return 1-sorted([p_y for p_y, y in self.p], reverse=True)[1]

    def credibility(self):
        """Credibility is an efficiency measure of a single prediction.
        Small credibility indicates an unusual example.

        Computes minimum :math:`\\mathit{eps}` that would result in an empty prediction set.
        :math:`\\mathit{eps} = \\text{max}(p_i)`

        Returns:
            float: Credibility :math:`\\mathit{eps}`.
        """
        return max(p_y for p_y, y in self.p)


class ConformalClassifier(ConformalPredictor):
    """Base class for conformal classifiers."""

    def __init__(self, nc_measure, mondrian=False):
        """Verify that the nonconformity measure can be used for classification."""
        assert isinstance(nc_measure, ClassNC), "Inappropriate nonconformity measure for classification"
        self.nc_measure = nc_measure
        self.mondrian = mondrian

    def __str__(self):
        return "{} ({})".format(self.__class__.__name__, self.nc_measure)

    def p_values(self, example):
        """Extending classes should implement this method to return a list of pairs (p-value, class)
        for a given example.

        Conformal classifier assigns an assumed class value to the given example and computes its nonconformity.
        P-value is the ratio of more nonconformal (stranger) instances that the given example.
        """
        raise NotImplementedError

    def predict(self, example, eps=None):
        """Compute a classification prediction object from p-values for a given example and significance level.

        Args:
            example (Instance): Orange row instance.
            eps (float): Default significance level (error rate).

        Returns:
            PredictionClass: Classification prediction object.
        """
        ps = self.p_values(example)
        return PredictionClass(ps, eps)

    def __call__(self, example, eps):
        """Compute predicted classes for a given example and significance level.

        Args:
            example (Instance): Orange row instance.
            eps (float): Significance level (error rate).

        Returns:
            List of predicted classes.
        """
        pred = self.predict(example)
        return pred.classes(eps)


class TransductiveClassifier(ConformalClassifier):
    """Transductive classification.

    Examples:
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> tcp = TransductiveClassifier(ProbabilityMargin(NaiveBayesLearner()), train)
        >>> print(tcp(test[0], 0.1))
    """

    def __init__(self, nc_measure, train=None, mondrian=False):
        """Initialize transductive classifier with a nonconformity measure and a training set.

        Fit the conformal classifier to the training set if present.

        Args:
            nc_measure (ClassNC): Classification nonconformity measure.
            train (Optional[Table]): Table of examples used as a training set.
            mondrian (bool): Use a mondrian setting for computing p-values.
        """
        super().__init__(nc_measure, mondrian=mondrian)
        if train is not None:
            self.fit(train)

    def fit(self, train):
        """Fit the conformal classifier to the training set and store the domain.

        Args:
            train (Optional[Table]): Table of examples used as a training set.
        """
        self.train = train
        self.domain = self.train.domain

    def p_values(self, example):
        """Compute p-values for every possible class.

        Transductive classifier appends the given example with an assumed class value to the training set
        and compares its nonconformity against all other instances.

        Args:
            example (Instance): Orange row instance.

        Returns:
            List of pairs (p-value, class)
        """
        ps = []
        temp = example.get_class()
        for yi, y in enumerate(self.domain.class_var.values):
            example.set_class(yi)
            data = Table(self.domain, np.vstack((self.train, example)))
            self.nc_measure.fit(data)
            scores = np.array([self.nc_measure.nonconformity(row) for row in data
                               if not self.mondrian or self.mondrian and row.get_class() == y])
            alpha, alpha_n = scores[:-1], scores[-1]
            p_y = sum(scores >= alpha_n) / len(scores)
            ps.append((p_y, y))
        example.set_class(temp)
        return ps


class InductiveClassifier(ConformalClassifier):
    """Inductive classification.

    Attributes:
        alpha: Nonconformity scores of the calibration instances. Computed by the :py:func:`fit` method.

    Examples:
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> train, calibrate = next(RandomSampler(train, 2, 1))
        >>> icp = InductiveClassifier(InverseProbability(LogisticRegressionLearner()), train, calibrate)
        >>> print(icp(test[0], 0.1))
    """

    def __init__(self, nc_measure, train=None, calibrate=None, mondrian=False):
        """Initialize inductive classifier with a nonconformity measure, training set and calibration set.
        If present, fit the conformal classifier to the training set and compute the nonconformity scores of
        calibration set.

        Args:
            nc_measure (ClassNC): Classification nonconformity measure.
            train (Optional[Table]): Table of examples used as a training set.
            calibrate (Optional[Table]): Table of examples used as a calibration set.
            mondrian (bool): Use a mondrian setting for computing p-values.
        """
        super().__init__(nc_measure, mondrian=mondrian)
        if train is not None and calibrate is not None:
            self.fit(train, calibrate)

    def fit(self, train, calibrate):
        """Fit the conformal classifier to the training set, compute and store nonconformity scores (:py:attr:`alpha`)
        on the calibration set and store the domain.

        Args:
            train (Optional[Table]): Table of examples used as a training set.
            calibrate (Optional[Table]): Table of examples used as a calibration set.
        """
        self.domain = train.domain
        self.calibrate = calibrate
        self.nc_measure.fit(train)
        self.alpha = np.array([self.nc_measure.nonconformity(inst) for inst in calibrate])

    def p_values(self, example):
        """Compute p-values for every possible class.

        Inductive classifier assigns an assumed class value to the given example and compares its nonconformity
        against all other instances in the calibration set.

        Args:
            example (Instance): Orange row instance.

        Returns:
            List of pairs (p-value, class)
        """
        classes = []
        ps = []
        temp = example.get_class()
        for yi, y in enumerate(self.domain.class_var.values):
            example.set_class(yi)
            alpha_n = self.nc_measure.nonconformity(example)
            if self.mondrian:
                alpha = np.array([a for a, cal in zip(self.alpha, self.calibrate) if cal.get_class() == y])
            else:
                alpha = self.alpha
            p_y = (sum(alpha >= alpha_n)+1) / (len(alpha)+1)
            ps.append((p_y, y))
        example.set_class(temp)
        return ps


class CrossClassifier(InductiveClassifier):
    """Cross classification.

    Examples:
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> ccp = CrossClassifier(InverseProbability(LogisticRegressionLearner()), 3, train)
        >>> print(ccp(test[0], 0.1))
    """

    def __init__(self, nc_measure, k, train=None, mondrian=False):
        """Initialize cross classifier with a nonconformity measure, number of folds and training set.
        If present, fit the conformal classifier to the training set.

        Args:
            nc_measure (ClassNC): Classification nonconformity measure.
            k (int): Number of folds.
            train (Optional[Table]): Table of examples used as a training set.
            mondrian (bool): Use a mondrian setting for computing p-values.
        """
        # store the unfitted nonconformity measure for making copies to fit on individual folds
        super().__init__(nc_measure, mondrian=mondrian)
        self.nc_measure_base = deepcopy(self.nc_measure)
        self.k = k
        if train is not None:
            self.fit(train)

    def fit(self, train):
        """Fit the cross classifier to the training set. Split the training set into k folds for use as
        training and calibration set with an inductive classifier. Concatenate the computed nonconformity scores
        and store them (:py:attr:`InductiveClassifier.alpha`).

        Args:
            train (Table): Table of examples used as a training set.
        """
        self.domain = train.domain
        self.calibrate = train
        self.nc_measure.fit(train)
        self.alpha = np.array([])
        for train_index, calibrate_index in KFold(self.k, shuffle=True).split(train):
            icp = InductiveClassifier(deepcopy(self.nc_measure_base), train[train_index], train[calibrate_index])
            self.alpha = np.concatenate((self.alpha, icp.alpha))


class LOOClassifier(CrossClassifier):
    """Leave-one-out classifier is a cross conformal classifier with the number of folds equal
    to the size of the training set.

    Examples:
        >>> train, test = next(LOOSampler(Table('iris')))
        >>> loocp = LOOClassifier(InverseProbability(LogisticRegressionLearner()), train)
        >>> print(loocp(test[0], 0.1))
    """

    def __init__(self, nc_measure, train=None, mondrian=False):
        super().__init__(nc_measure, 0, train, mondrian)

    def fit(self, train):
        self.k = len(train)
        super().fit(train)




class ConformalPredictor:
    """Base class for conformal predictors."""

    def __call__(self, example, eps):
        """Extending classes should implement this method to return predicted values
        for a given example and significance level."""
        raise NotImplementedError

    def predict(self, example, eps):
        """Extending classes should implement this method to return a prediction object.
        for a given example and significance level."""
        raise NotImplementedError



from orangecontrib.conformal import base
from orangecontrib.conformal import classification
from orangecontrib.conformal import evaluation
from orangecontrib.conformal import nonconformity
from orangecontrib.conformal import regression

__version__ = "1.0.1"
